<?php defined('SYSPATH') OR die('No direct access allowed.');

class Geomap_Model extends Model
{
	/**
	 * A map of node types to their fields.
	 */
	private static $type_keys = array(
		'link' => array(
			'id',
			'id1',
			'id2',
			'description'
		),
		'location' => array(
			'id',
			'point',
			'label',
			'address',
			'description'
		)
	);

	/**
	 * A map of object types to their required fields. There is no room for optional fields.
	 * The only exception is the 'type' key that is magical.
	 * Any node that doesn't fit the bill will be silently ignored.
	 */
	private static $obj_keys = array(
		'host' => array(
			'name'
		),
		'service' => array(
			'host',
			'description'
		),
		'hostgroup' => array(
			'name'
		),
		'servicegroup' => array(
			'name'
		)
	);

	/**
	 * Keep an object around, as it is expensive to create more
	 */

	private $locations = false; /**< ID indexed array of locations */
	private $links = false; /**< ID indexed array of links */
	private $viewpoints = false; /**< Doesn't do anything, does it? Called by the frontend, though */

	public function __construct() {
		parent::__construct();
	}

	/**
	 * Load an XML file.
	 *
	 * @param $type Which xml file type to load
	 * @return false on error, true otherwise
	 */
	private function read_nodes($type)
	{
		$plural = $type . 's';
		switch ($type) {
		 case 'location':
		 case 'link':
		 case 'viewpoint':
			break;
		 default:
			return false;
		}

		$xml = @simplexml_load_file(Kohana::config('geomap.config_path') . $plural . '.xml');
		$this->$plural = array();
		if (!isset(self::$type_keys[$type]))
			return false;
		foreach ($xml->$type as $node)
		{
			$content = array();

			// Remove illegal attributes
			// and make sure all valid ones are at least set
			foreach ($node->attributes() as $attr => $value) {
				if (!in_array($attr, self::$type_keys[$type]))
					continue; // Illegal attribute, skip
				$content[$attr] = (string)$value;
			}

			foreach (self::$type_keys[$type] as $valid) {
				if (!isset($content[$valid]))
					$content[$valid] = '';
			}

			/* Note: there should be only one child of a node,
				 but it is required to use foreach with children() */
			foreach ($node->children() as $subnode)
			{
				$node_type = $subnode->getName();
				if (!isset(self::$obj_keys[$node_type]))
					continue 2;
				$content['object']['type'] = $node_type;

				foreach ($subnode->attributes() as $attr => $value) {
					if (!in_array($attr, self::$obj_keys[$node_type]))
						continue;
					$content['object'][$attr] = (string)$value;
				}

				foreach (self::$obj_keys[$node_type] as $key) {
					if (!isset($content['object'][$key]))
						continue 3;
				}
			}

			$this->{$plural}[$content['id']] = $content;
		}

		return true;
	}

	private function create_node($type, $content)
	{
		$plural = $type . 's';
		switch ($type) {
		 case 'location':
		 case 'link':
			break;
		 default:
			return false;
		}

		$valid_content = array();
		if (!isset(self::$type_keys[$type]))
			return false;
		foreach (self::$type_keys[$type] as $attr) {
			if (isset($content[$attr]))
				$valid_content[$attr] = $content[$attr];
			else
				$valid_content[$attr] = '';
		}
		if (isset($content['object'])) {
			if (!isset($content['object']['type']) || !isset(self::$obj_keys[$content['object']['type']]))
				return false;
			$valid_content['object']['type'] = $content['object']['type'];
			foreach (self::$obj_keys[$content['object']['type']] as $attr) {
				if (!isset($content['object'][$attr]))
					return false;
				$valid_content['object'][$attr] = $content['object'][$attr];
			}
			if($this->is_allowed($valid_content['object']) === FALSE)
				return false;
		}

		$this->{$plural}[$valid_content['id']] = $valid_content;
		return true;
	}

	private function write_nodes($type) {
		$plural = $type . 's';
		$xml = new SimpleXMLElement("<{$type}s></{$type}s>");
		foreach ($this->$plural as $obj) {
			$node = $xml->addChild($type);
			foreach ($obj as $attr => $value)
			{
				if ($attr == 'object' && isset($value['type'])) {
					$subnode = $node->addChild($value['type']);
					foreach ($value as $objattr => $objvalue)
						if ($objattr !== 'type' && !empty($objvalue))
							$subnode->addAttribute($objattr, $objvalue);
				}
				else if (!empty($value)) {
					$node->addAttribute($attr, $value);
				}
			}
		}

		if (@file_put_contents(Kohana::config('geomap.config_path') . $plural .'.xml', $xml->asXML()) === FALSE)
			throw new Exception("Could not write {$plural}.xml");

		return true;
	}

	private function update_node($type, $content)
	{
		$this->delete_node($type, $content['id']);
		$this->create_node($type, $content);
	}

	private function delete_node($type, $id)
	{
		$plural = $type.'s';
		switch ($type) {
		 case 'location':
		 case 'link':
			break;
		 default:
			return false;
		}

		if (!isset($this->{$plural}[$id]))
			return false;

		$obj = $this->{$plural}[$id];

		if (isset($obj['object']) && !$this->is_allowed($obj['object']))
			return false;

		unset($this->{$plural}[$id]);
		return true;
	}

	private function is_allowed($object) {
		switch ($object['type'])
		{
			case 'host':
				return false !== HostPool_Model::fetch_by_key($object['name']);
			case 'service':
				return false !== ServicePool_Model::fetch_by_key("{$object['host']};{$object['description']}");
			case 'hostgroup':
				return false !== HostGroupPool_Model::fetch_by_key($object['name']);
			case 'servicegroup':
				return false !== ServiceGroupPool_Model::fetch_by_key($object['name']);
		}
		return FALSE;
	}

	/**
	 * Get the state for the specified object
	 *
	 * @return the state (an integer or string), or null if the object wasn't found
	 */
	private function get_state($object)
	{
		switch ($object['type'])
		{
			case 'host':
				$state = HostPool_Model::all()->reduce_by('name', $object['name'], '=')->one('state');
				if(!$state) {
					return null;
				}
				return $state->get_state();
			case 'service':
				$state = ServicePool_Model::all()
					->reduce_by('host.name', $object['host'], '=')
					->reduce_by('description', $object['description'], '=')
					->one('state');
				if(!$state) {
					return null;
				}
				return $state->get_state();

			case 'hostgroup':
				$state = HostGroupPool_Model::all()
					->reduce_by('name', $object['name'], '=')
					->one('worst_host_state');
				if(!$state) {
					return null;
				}
				return $state->get_worst_host_state();

			case 'servicegroup':
				$state = ServiceGroupPool_Model::all()
					->reduce_by('name', $object['name'], '=')
					->one('worst_service_state');
				if(!$state) {
					return null;
				}
				return $state->get_worst_service_state();
		}
		return null;
	}

	public function get_settings()
	{
		if (($xml = @simplexml_load_file(Kohana::config('geomap.config_path') . 'settings.xml')) === FALSE)
			throw new Exception('Could not read settings.xml');

		$settings = array();

		foreach ($xml->attributes() as $attr => $value)
			$settings[$attr] = (string)$value;

		if (isset($settings['provider']))
		{
			if (!in_array($settings['provider'], array('google', 'osm', 'local'))
				|| $settings['provider'] == 'osm' && !isset($settings['cloudmadeKey'])
				|| $settings['provider'] == 'local' && !isset($settings['localMapsUrl']))
			{
				throw new Exception('Invalid settings.xml');
			}
		}
		else
			$settings['provider'] = 'google';

		return $settings;
	}

	public function save_settings(array $data) {
		if (($xml = @simplexml_load_file(Kohana::config('geomap.config_path') . 'settings.xml')) === FALSE)
			throw new Exception('Could not read settings.xml');

		foreach ($xml->attributes() as $key => $val) {
			if (isset($data[$key])) {
				if ($key === 'cloudmadeKey' && ($data['provider'] === 'google' || $data['provider'] === 'local'))
					return FALSE;
				if ($key === 'localMapsUrl' && ($data['provider'] === 'google' || $data['provider'] === 'osm'))
					return FALSE;
				if($key === 'cloudmadeKey' && strlen($data[$key]) !== 32) {
					return FALSE;
				}
				$xml[$key] = $data[$key];
			}
		}

		if(@file_put_contents(Kohana::config('geomap.config_path') . 'settings.xml', $xml->asXML()) === FALSE) {
			throw new Exception('Could not write settings.xml');
		}
		return TRUE;

	}

	/**
	 * Return all locations, filtered and with state added
	 */
	public function get_locations()
	{
		if ($this->locations === false)
			$this->read_nodes('location');
		if ($this->locations === false)
			return false;

		$locations = array();
		foreach ($this->locations as $key => $location)
		{
			if (isset($location['object'])) {
				if($this->is_allowed($location['object'])) {
					$location['object']['state'] = $this->get_state($location['object']);
					if ($location['object']['state'] !== null)
						$locations[] = $location;
				}
			}
			else
				$locations[] = $location;
		}
		return $locations;
	}

	/**
	 * Return all links, filtered and with state added
	 */
	public function get_links()
	{
		if ($this->links === false)
			$this->read_nodes('link');
		if ($this->links === false)
			return false;

		$links = array();
		foreach ($this->links as $key => $link)
		{
			if (isset($link['object'])) {
				if($this->is_allowed($link['object'])) {
					$link['object']['state'] = $this->get_state($link['object']);
					if ($link['object']['state'] !== null)
						$links[] = $link;
				}
			}
			else
				$links[] = $link;
		}
		$links = array( 'items' => $links );

		return $links;
	}

	public function get_viewpoints()
	{
		if ($this->viewpoints === false)
			$this->read_nodes('viewpoint');
		return $this->viewpoints;
	}

	public function add_location($location)
	{
		if ($this->locations === false)
			$this->read_nodes('location');

		$location['id'] = uniqid('', true);
		$this->create_node('location', $location);

		$this->write_nodes('location');

		return $location['id'];
	}

	public function add_link($link)
	{
		if ($this->links === false)
			$this->read_nodes('link');

		$link['id'] = uniqid('', true);
		$this->create_node('link', $link);

		$this->write_nodes('link');

		return $link['id'];
	}

	public function edit_location($location)
	{
		if ($this->locations === false)
			$this->read_nodes('location');

		$this->update_node('location', $location);
		$this->write_nodes('location');
	}

	public function edit_link($link)
	{
		if ($this->links === false)
			$this->read_nodes('link');

		$this->update_node('link', $link);
		$this->write_nodes('link');
	}

	public function remove_location($location)
	{
		if ($this->locations === false)
			$this->read_nodes('location');

		$this->delete_node('location', $location['id']);
		$this->write_nodes('location');
	}

	public function remove_link($link)
	{
		if ($this->links === false)
			$this->read_nodes('link');

		$this->delete_node('link', $link['id']);
		$this->write_nodes('link');
	}
}
